// EGO controller
   /*!
   * \addtogroup g_ego_nbo2 Narrow Band O2 Sensor Routines
   */
/*@{*/

#include "global.h"
#include "ego.h"
#include "adc.h"
#include "helpers.h"
#include "fuelcalc.h"
#include "eventqueue.h"
#include "ve.h"
#include "storage.h"
#include "pid.h"
#include "wbo2.h"
#include "arithm.c"
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

uint8_t ego_history[HIST_LENGTH];
uint8_t ego_idx;
uint16_t ego_corr_filt;	// filtered value, sum of HIST_LENGTH samples in fact
uint16_t target_afr;

struct pid_t ego_pid;

// initialize the ego warmup counter
// a value of ff defaults to 60 seconds
void init_ego(void) {
	etime.ego_warmup_cnt = config.ego_warmup;
	pid_init(&ego_pid, (struct pid_parameters_t*) &config.ego_pid_kp);
}


// run this every second
void ego_nbo2_ready(void) {
// TODO: measure the o2 sensor output too

	if (engine.status & _BV(running)) {  // engine is running,
		if (engine.status_ext & _BV(o2_not_ready)) { // and the O2 isn't ready
			if (etime.ego_warmup_cnt > 0) {
				if(--etime.ego_warmup_cnt == 0){	// wait some time
					cli();
					engine.status_ext &= ~_BV(o2_not_ready);  // enable the O2
					sei();
				}
			}
		}
	} else {
		cli();
		engine.status_ext |= _BV(o2_not_ready);  // disable the O2
		sei();

		if (etime.ego_warmup_cnt <= (config.ego_warmup - EGO_COOL)) {
			etime.ego_warmup_cnt += EGO_COOL;  // let the O2 "cool" fast
		}
	}
}


// this table is generated by bin/lambda.pl
// elements are:  afr*1024

uint16_t PROGMEM lambdacorr_to_afr[] = 
{19268, 17840, 16610, 15538, 14597, 13763, 
13019, 12351, 11749, 11202, 10704, 10249,
9830, 9445, 9088, 8758, 8450};

uint16_t calc_target_afr(void) {
/* 
   lambda table can modify the VE table in the range  78.1% ... 177.7%
   It is biased at 0x38 (afr=14.7) with 0x00 (afr=18.8) and 0xff (afr=5.6)

   lambdacorr_to_afr table consists of 17 evenly spaced elements
*/

	uint8_t idx, offset;
	uint16_t a, b, res;

	idx = corr.lambda;
	offset = idx & 0x0f;
	idx >>= 4;

	a = pgm_read_word(&lambdacorr_to_afr[idx]);
	b = pgm_read_word(&lambdacorr_to_afr[idx+1]);

	// table lambda_to_afr[]is decreasing, therefore this works
	res = a - (mult16_8(a - b, offset) >> 4);
	target_afr = res; // store it globally somewhere

//	return res & 0x7fff;
	return res;  // table elements is <= 15 bits
}

/* symmetric railing */
static inline uint8_t sym_rail(int16_t candidate)
{
	uint8_t upper, lower, cl;
	cl = config.ego_lean_limit;
	if (cl >= EGOCORR_BASE) {
		upper = 0xff;
		lower = 0x00;
	} else {
		upper = EGOCORR_BASE + cl;
		lower = EGOCORR_BASE - cl;
	}

	// *signed* limit
	return limit_i16(candidate, (uint16_t)upper, (uint16_t)lower);
}

/* asymmetric railing.
Note: candidate was uint8_t originally. (??)
*/
static inline uint8_t asym_rail(uint16_t candidate)
{
	uint8_t cl, higher_byte, candi;
	
	// this should never be called with "negative" numbers, but let's be cautios for now
	higher_byte = candidate >> 8;
	if(higher_byte){
		if(higher_byte < 0x80){ 
			candi = 0xff;
		} else { // it was probably ment negative, zzz 	
			candi = 0;
		}
	} else {
		candi = (uint8_t) candidate; // higher_byte==0, so lower byte is fine
	}

	// candi is now candidate railed to 0..255 limits
	// note that this also protects against stupid user config, when ...limit > 127
	if(candi >= EGOCORR_BASE){ // enrich
		cl = config.ego_rich_limit;
		if((candi - EGOCORR_BASE) > cl)
			return EGOCORR_BASE + cl;				
	} else { // enlean
		cl = config.ego_lean_limit;
		if((EGOCORR_BASE - candi) > cl)
			return EGOCORR_BASE - cl;
	}
	
	return candi;
	
}	/* asym_rail() */

#ifdef NBO2_PID_EGO_EXPERIMENT
/*
PID ego controller: this shouldn't be needed ever.
WBO2 should use the incredible_ego while 
NBO2 is OK with the bangbang
 */
static inline void pid_ego(void)
{
	int16_t target, current;
#ifdef WBO2
	if (config.ego_conf & _BV(WBO2_SENSOR)) {
		target = calc_target_afr();
		current = wbo2.current_afr;
	} else 
#endif
	{
		target = config.ego_target<<7;
		current = sensors8(EGO_1)<<7;
	}

	// check if the current afr is within target +/- window
	if ((current <= (target+config.ego_pid_window)) && 
	    (current >= (target-config.ego_pid_window))) {
		; // do nothing, we are on target already, maybe return ?
	} else {
		uint16_t ego_candidate;

		pid_set_target(&ego_pid, target);
		ego_candidate = add8_16idiv16(EGOCORR_BASE, -pid15(&ego_pid, current));
		// TODO: integral "anti-windup"

		if (!(config.ego_conf & _BV(WBO2_SENSOR))) //NBO2 sensor has reversed slope
			ego_candidate = 2*EGOCORR_BASE-1 - ego_candidate;

		// TODO: make sure ego_candidate is non-negative!
		corr.ego = asym_rail(ego_candidate);
	}
} /* pid_ego() */

#endif // NBO2_PID_EGO_EXPERIMENT

/* The real thing: only meaningful with real measured AFR data
 ego correction algorithm: corr.ego *= AFRmeasured / AFRtarget
 reference data: corr.ego was "to richen": 154==0x80+26 (26/256 == +10% fuel)
 corr.lambda (==107) wants (1.2=(200+107)/256 =>  AFR=14.7/1.2)
 but AFR shows only slightly rich: 14.7 / 1.05 (*1024) == 14336
 
 in this case
 ego_candidate = corr.ego *1.2 / 1.05 ; which is 322 (-128)
 reference values in []
*/
#ifdef WBO2
static inline void incredible_ego(void)
{
	uint16_t newcorrego, corrold, t16, t17, afr, lambdacorr;
	uint8_t ce, ego_candidate, speed, speed_limit, rpm, speed256;

	afr = wbo2.current_afr;	// [14336]
	// do we want deadband ? (check if the current afr is within target +/- window)
	
	// 1/AFRtarget is already there, called lambdacorr (== 14.7/AFRtarget*256-200)! 
	lambdacorr = (uint16_t)corr.lambda + LAMBDACORR_OFFSET;	// [307]

	ce = corr.ego;	// [154]
	corrold = (uint16_t)ce + EGOCORR_BASE; // [282]
	t16 = mult16_16(lambdacorr, 1115) >> 8; // [1337], max 455 * 1115 /256 == 1982
	t17 = mult16_16(corrold, afr)>>8; // [15792] max 383 * 2*1024*14.7 / 256 == 45040
	newcorrego = mult16_16(t16, t17)>>16;	// [322]
	if(newcorrego > EGOCORR_BASE){
		newcorrego -= EGOCORR_BASE;	// [194]
		if((newcorrego>>8) == 0)
			ego_candidate =	(uint8_t)newcorrego;	// [194]
		else 
			ego_candidate = 0xff;		
	} else
		ego_candidate = 0;
 
	// Let's slow down the controller a bit using weighting:
	// ego_candidate = (newcorrego + corrold)>>1;

	// slower at lower RPM-s
	rpm = engine.rpm;

	if(rpm >= 64)
		speed = 0xff;
	else 
		speed = rpm << 2;

	// reusing the config variable to limit speed (suggested 0xc0)
	// TODO: confirm new strategy:
	// now synchronized to calc_afr (check wbo2.h WBO2_AFR_SAMPLES)
	// so calling frequency does NOT depend on processor speed and load.
	speed_limit = config.ego_pid_kp;
	if (speed > speed_limit){
		speed = speed_limit;
	} else {
		if (speed == 0)	// this cannot be allowed
			speed = 1;
	}

	speed256 = 0 - speed;	// 256 - in fact
	
	// no overflows possible:
	ego_candidate = (speed * ego_candidate)>>8;
	ego_candidate += (speed256 * ce)>>8;
	corr.ego = asym_rail(ego_candidate);
	
/* 
//	The lambdacorr table should be documented more 
// (stoich is 0x54 not 0x80 !, because 172+0x54 ==256)

ToThink:
If AFRtarget is very low, slowly downscale corr towards 1.0 (128) until 
AFRmeasured gets into the measurable range.. but not lean, of course... 
*/
} /* incredible_ego */
#endif
/* 
 megasquirt compatible BANG-BANG ego controller
 This controller is only for NBO2 sensors. 
*/
static inline void bangbang_ego(void)
{
	uint8_t target, ego_volt;

	target = config.ego_target;

	ego_volt = ego_voltage();


	if (ego_volt == target) {  // got the exact ego voltage (impressive!)
		return; // do nothing
	} else {
		uint8_t o2_is_lean;
		int16_t ego_candidate;
	
		o2_is_lean = (ego_volt < target ? true : false);
/* 		WBO2 uses incredible, therefore never gets here.
		Who would ever want to use the bangbang controller with a WBO2?
		if(WBO2_SENSOR_CONFIGURED) //wideband sensor has reversed slope
			o2_is_lean ^= true;	// true becomes false, false becomes true
*/
		//  config.ego_delta could reuse the PID parameters, never used at the same time
		if (o2_is_lean)      //afr is lean
			ego_candidate = corr.ego + config.ego_delta;
		else      //afr is rich
			ego_candidate = corr.ego - config.ego_delta;

		// megatune has the same rich/lean limit, take care!
		corr.ego = sym_rail(ego_candidate);
	}
} /* bangbang_ego() */

static inline void filt_ego(void)
{
	uint8_t *p, ce;
	
//			for (i=0; i<HIST_LENGTH; i++)
//				mean += ego_history[i];

	// store the last samples and calculate the mean ego correction
	// this is just for lcd-display purposes
	if (ego_idx >= HIST_LENGTH)
		ego_idx = 0;

	// we don't bother to initialize to stoich (are all 0 instead)
	// since only for display, not feedback
	p = &ego_history[ego_idx];
	ego_corr_filt -= *p;
	ce = corr.ego;
	ego_corr_filt += ce;
	*p = ce;
	ego_idx++;
}
/***************************************************************************/
/*** Exhaust gas oxygen sensor measurement section                       ***/
/***************************************************************************/
void o2(void) {

	// some reasons not to enter closed loop...
	if ( (config.ego_delta == 0) ||
	     (engine.coolant < config.ego_coolant) ||
	     (engine.status & (_BV(tpsaen) | _BV(tpsden))) ||
	     (engine.rpm <= config.ego_minrpm) ||
	     (engine.rpm >= config.ego_maxrpm) ||
	     (engine_kpa8() >= config.ego_maxmap) ||
	     (corr.tps_overrun & _BV(OVERRUN_ACTIVE)) ||
	     ((!(WBO2_SENSOR_CONFIGURED)) && (engine.tps >= config.ego_maxtps)) ||
	     (engine.status_ext & (_BV(o2_not_ready) | _BV(als_active)))  // skip the first n sec and disable while bangbanging
		) {

		corr.ego = EGOCORR_BASE;  // this should converge instead of an abrupt adjust 
//		ego_corr_average = EGOCORR_BASE;
		etime.egocount = config.ego_lag;
		
		cli();
		engine.status_ext &= ~_BV(o2_closed_loop);
		sei();
		return;
	} 
	
	cli();
	engine.status_ext |= _BV(o2_closed_loop);
	sei();
		
	if (etime.egocount != 0)
		return;

	etime.egocount = config.ego_lag;
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED)
		incredible_ego();
	else
#endif
#ifdef NBO2_PID_EGO_EXPERIMENT
		if(config.ego_conf & _BV(PID_EGO_CONTROL)) {
			pid_ego();
		} else	
#endif	// NBO2_PID_EGO_EXPERIMENT
		// megatune compatible ego controller
			bangbang_ego();

	filt_ego();
}

 /*@}*/

