/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
\*********************************************************/

   /*!
   * \addtogroup g_idle_air Idle Air Control Routines
   */
/*@{*/

#include "global.h"
#include "iac.h"
#include "helpers.h"
#include "adc.h"
#include "ve.h"
#include "fuelcalc.h"
#include "benchmark.h"

#include <inttypes.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "arithm.c"
#include "lbus.c"

#include "softpwm.c"
#include "injector.c"

struct softpwm_t idle_pwm_soft;

struct iac_t iac;

uint8_t idle_rpm;  // the target_rpm
uint8_t iac_pos;   // the calculated absolute iac position

#ifdef SIMPLE_IAC
#warning "Don't use this idle controller, the other is much better!"

/***************************************************************************/
/*** calc iac pos                                                        ***/
/***************************************************************************/
// The unit of all rpm variables are [actual_rpm/10]
void calc_iac_pos(void) {
	static uint8_t last_iac_pos;

	if (!(engine.status & _BV(running))) { // engine ain't running
		uint8_t iac_pos;

		cli();
		engine.status &= ~_BV(idle);
		sei();

		// park the iac at the starting position
		idle_rpm = config.iac_cold_rpm;
		iac_pos = linear_interp( COLD_TEMP, WARM_TEMP, config.iac_cold_start_pos, config.iac_warm_start_pos, engine.coolant );
		if( iac_pos > config.iac_max_steps ) iac_pos = config.iac_max_steps;
		last_iac_pos = iac_pos;
		set_iac_pos(iac_pos);

	} else {  // either cranking or running
					
		if (engine.status & _BV(crank)) {  // no movement of iac while cranking
			cli();
			engine.status &= ~_BV(idle);
			sei();

		} else {

			if (engine.tps <= config.iac_tps_thres) {
				uint16_t current_rpm;
				
				cli(); // everything that can go wrong, will go wrong
				current_rpm = engine.rpm_vhr;
				sei(); // ... unless we use cli/sei :-)

// (REAL rpm_vhr=900, engine.rpm==9, current_rpm=90 or idle_rpm==90)
				current_rpm = div10(current_rpm);

				// the FAST_IDLE_THRES avoids idle_rpm switching back and forth when coolant ~ fastidle
				if (engine.coolant >= config.iac_fast_idle_temp) {
					idle_rpm = config.iac_warm_rpm;
				} else if (engine.coolant < (config.iac_fast_idle_temp - FAST_IDLE_THRES)) {
					idle_rpm = config.iac_cold_rpm;
				}


				if (!(engine.status & _BV(idle))) {  // just entered idle mode
					cli();
					engine.status |= _BV(idle);
					sei();

					// use the iac position from last time
					iac_pos = limit_u8(last_iac_pos, config.iac_max_steps, 0);
					set_iac_pos(iac_pos);

				} else if (!(iac.status & _BV(busy))) { // if not busy moving the iac

					static uint8_t fine_count, coarse_count;
					static uint8_t sync_count;
					int8_t adjust=0;

					if (++coarse_count >= config.iac[COARSE_PERIOD]) { // how often can the iac be moved for coarse adjustments
						coarse_count=0;

						if (current_rpm > (idle_rpm + config.iac[COARSE_DEV])) { // going way too fast
							adjust = -(config.iac[ADJUST_COARSE] & 0x0F);
							sync_count=0;
							fine_count=0;
						} else if (current_rpm < (idle_rpm - config.iac[COARSE_DEV])) {  // going way too slow
							adjust = config.iac[ADJUST_COARSE] >> 4;
							sync_count=0;
							fine_count=0;
						} else {
							adjust = 0;  // no adjustments here
						}
					}
					
					if (adjust == 0) {  // if any coarse adjustments, then don't mess with them
						if (++fine_count >= config.iac[FINE_PERIOD]) { // how often can the iac be moved for fine adjustments
							fine_count=0;
						
							if (current_rpm > (idle_rpm + config.iac[FINE_DEV])) { // going a little too fast
								adjust = -(config.iac[ADJUST_FINE] & 0x0F);
								
								// resync the iac if it has drifted.
								// will only be resync'ed when there is a small deviation from target-rpm
								// and if that deviation can't be eliminated otherwise.
								if (iac.position == 0) {
									if (sync_count < 255) sync_count++;
									if (sync_count >= config.iac[RESYNC]) { // time for resync'ing the iac
										sync_count=0;
										cli();
										if (iac.position==0) iac.position++;  // cheat the iac, now it thinks it is at position 1
										sei();
									}
								}
							} else if (current_rpm < (idle_rpm - config.iac[FINE_DEV])) {  // going a little too slow
								adjust = config.iac[ADJUST_FINE] >> 4;
								sync_count=0;
							} else { // hey, then we must have reached the right rpm, hurray!
								adjust = 0; // no adjustments here
								sync_count=0;
							}
						}
					}
					// limit the iac adjustment
					// we use the iac.position as a floating reference
					iac_pos = limit_i16(iac.position + adjust, config.iac_max_steps, 0);
					set_iac_pos(iac_pos);
				}
				
			} else {  // not in idle-mode

				// remember last iac position
				if (iac.status & _BV(idle)) {
					last_iac_pos = iac_pos;
					cli();
					engine.status &= ~_BV(idle);
					sei();
				}
			}
		}
	}
}
#endif




#ifdef PID_IAC

/*
  Freakin Fantastic Idle Control [FFIC (tm)]
  proportional change in ignition advance as function of deviance from target rpm.
 */
uint8_t idle_advance_adjust(uint8_t adv) {
	uint8_t new_adv;
//TODO: _BV(advance_control) is stealing a bit that is already used
//      for specifying the stepper speed. Find another bit!
// in the meantime,
//  set odd msec stepper speed if you want advance_control ON,
// and even msec stepper speed if you want advance_control OFF
// or set odd msec and use other variables to limit ignition advance adjust, eg.:
// iac_ign_advance_limit=00
// iac_ign_retard_limit=00
	if ((config.iac_conf & _BV(advance_control)) && (engine.status & _BV(idle))) {
		uint16_t current_rpm, target_rpm;
		uint8_t limit;

		current_rpm = engine.rpm_vhr;
		target_rpm = idle_rpm * 10;
		
		if (current_rpm < target_rpm) {  // going too slow; more advance
			uint16_t t, adjust;
			t = target_rpm - current_rpm;

			adjust = mult16_8(t, config.iac_ign_advance_change) >> 8;
			limit = config.iac_ign_advance_limit;

			if (adjust > limit)
				adjust = limit;

			t = (uint16_t)adv + adjust;
			if (t>>8) { // this cannot be healthy
				new_adv = 0xff;
			} else {
				new_adv = (uint8_t)t;
			}

		} else {  // going too fast; less advance
			uint16_t t, adjust;
			t = current_rpm - target_rpm;

			adjust = mult16_8(t, config.iac_ign_retard_change) >> 8;
			limit = config.iac_ign_retard_limit;

			if (adjust > limit)
				adjust = limit;

			t = (uint16_t)adv - adjust;
			if (t>>8) {
				new_adv = 0;
			} else {
				new_adv = (uint8_t)t;
			}
		}

	} else {
		new_adv = adv;
	}

	iac.adv_adjust = new_adv - adv;
	return new_adv;
}

/***************************************************************************/
/*** calc iac pos                                                        ***/
/***************************************************************************/
// The unit of internal rpm variables is [1 rpm] unless otherwise noted
// The caller is advised to only call if(config.iac_conf & _BV(precise_idle))
void calc_iac_pos(void) {
	uint16_t current_rpm;
	static uint16_t last_rpm;
	static uint8_t afterstart_duration; // timer for controlling the extra air at start
	static uint8_t afterstart_rpm;
	static uint8_t target_rpm;

	if ((!(engine.status & _BV(running))) || (engine.status & _BV(crank))) { // engine ain't running or in cranking mode

		cli();
		engine.status &= ~_BV(idle);
		sei();

		iac.error = 0;
		iac.integral = (uint16_t)config.iac_integral_limit_inc << INTEGRAL_LIMIT_SCALE;
		iac.rpmdot = 0;
		iac.pid = 0;
		iac.pid_p = 0;
		iac.pid_i = 0;
		iac.pid_d = 0;
		iac.resync = 10 * config.iac_overclose_interval;
		iac.adv_adjust = 0;

		// park the iac at the starting position
		afterstart_duration = config.iac_afterstart_duration;  // initialize iac warmup counter
		afterstart_rpm = config.iac_afterstart_rpm;
		target_rpm = idle_rpm = config.iac_cold_rpm + afterstart_rpm;

		// we have a seperate starting position to help getting air into the cylinder
		iac_pos = linear_interp( COLD_TEMP, WARM_TEMP, config.iac_cold_start_pos, config.iac_warm_start_pos, engine.coolant );
		if( iac_pos > config.iac_max_steps ) iac_pos = config.iac_max_steps;
		set_iac_pos(iac_pos);

	} else {  // running

		if (engine.tps <= config.iac_tps_thres) {
			int16_t pos;

			if (!(engine.status & _BV(idle))) {  // just entered idle mode
				cli();
				engine.status |= _BV(idle);
				sei();
			}

			current_rpm = engine.rpm_vhr;  // only written from userspace, no need for cli/sei

// engine.rpm==9, REAL_RPM == rpm_vhr() == 900 
// for target config.idle_rpm == 90 == 0x5a

			idle_rpm = linear_interp(config.iac_cold_idle_temp,
							 config.iac_warm_idle_temp,
							 config.iac_cold_rpm,
							 config.iac_warm_rpm,
							 engine.coolant);

				
			// "Afterstart Air Enrichment"
			// this can be used for kickstarting the alternator, or just to get the engine running :-)
			// Add some extra 'afterstart' rpm to idle_rpm

			if ( afterstart_rpm == 0 ) {
				pos = 0;
			} else {
				if (afterstart_duration) afterstart_duration--;
				else if (afterstart_rpm) afterstart_rpm--;
				
				pos = (config.iac_afterstart_steps * afterstart_rpm) >> 4;  // steps per 160 rpm
			}

			idle_rpm += afterstart_rpm;

			// error term
			// a positive error means the engine is running slower than it is supposed to do
			iac.error = (idle_rpm*10) - current_rpm;
      
			// asymmetric behavior. If rpm is low, increase rpm fast
			if (config.iac_pid_conf & _BV(pid_asymmetric)) {
				if (iac.error > 0)
					iac.error += iac.error;
			}

			// check if the error is large enough to allow adjustments.
			// if the deadband is too small, the iac will be adjusting all the time (not recommend!)
			if (abs(iac.error) >= config.iac_deadband) {
				// Differential change of rpm
				// A positive change in rpm results in negative d(rpm). This is caused by the way 'error' is defined
				iac.rpmdot = last_rpm - current_rpm;

				if (abs(iac.error) >= config.iac_integral_deadband) {
				// integral windup
				if ((iac_pos == 0) && (iac.error < 0)) {
					// negative error, but iac can't move any lower
					;
				} else if ((iac_pos == config.iac_max_steps) && (iac.error > 0)) {
					// positive error, but iac can't more any higher
					;
				} else {
					int16_t i16;
					int32_t i;
					int16_t inc_limit, dec_limit;

					// integrate error term
					// scale the amount to modify the integral variable
					i = (int32_t)iac.integral + mult16i_8_div256(iac.error, config.iac_integral_speed);

					// handle overflow
					i16 = i>>8;
					if (i16 > 127)
						iac.integral = 32767;
					else if (i16 < -128)
						iac.integral = -32768;
					else
						iac.integral = i;

					// yes please, we really want to limit the integrator
					// The decrease-limit should be as small as possible, else the I-term will
					// suffer from the wind-up. The increase-limit can be higher to allow the 
					// target rpm to be reached at hot start

					inc_limit = (uint16_t)config.iac_integral_limit_inc << INTEGRAL_LIMIT_SCALE;
					dec_limit = -((uint16_t)config.iac_integral_limit_dec << INTEGRAL_LIMIT_SCALE);

					iac.integral = limit_i16(iac.integral, inc_limit, dec_limit);
				}
				}

				iac.pid_p = mult16i_8_div256(iac.error, config.iac_kp);
				iac.pid_i = mult16i_8_div256(iac.integral, config.iac_ki);
				iac.pid_d = mult16i_8_div256(iac.rpmdot, config.iac_kd);

				// lets hope this doesn't overflow
				// with sufficiently small (iac_kp + iac_ki + iac_kd <= 256) values will never overflow
				// but in practice iac.error and iac.rpmdot is also small (compared to 32767 RPM)
				// and iac_integral_limit helps too => no worries
				iac.pid = iac.pid_p + iac.pid_i + iac.pid_d;

				// divide by 16 to allow larger constants in the above 
				// iac.pid = mult16i_8_div256(iac.pid, 16);
				iac.pid = div16i_exp2(iac.pid, 4);
				
				// final computed iac position change
				// a signed word should be enough for this
				pos += iac.pid;

				// now lookup the predefined "the-iac-should-be-around-this-position-at-this-temperature" position
				pos += linear_interp(config.warmup_clt_range[clt_idx.lindex],
						     config.warmup_clt_range[clt_idx.uindex],
						     config.iac_ref_pos[clt_idx.lindex],
						     config.iac_ref_pos[clt_idx.uindex],
						     engine.coolant);

				// resync the iac to correct any drift
				if ( iac.position == 0 ) {
					if (iac.resync != 0) {
						iac.resync--;
					} else {
						iac.resync = 10 * config.iac_overclose_interval;
						iac.position=1;  // cheat the iac, now it thinks it is at position 1
						// no need for cli/sei, iac is moved in userspace
//#ifdef BENCHMARK
//						benchstats.iac_overclose++;
//#endif
					}
				}


				// limit the iac adjustment
				iac_pos = limit_i16(pos, config.iac_max_steps, 0);

				set_iac_pos(iac_pos);

			} else if (abs(iac.adv_adjust) >= config.iac_ign_threshold ) {
				pos = iac.position;

				if (iac.error > 0) {
					pos++;
				} else {
					pos--;
				}
				
				iac_pos = limit_i16(pos, config.iac_max_steps, 0);
				set_iac_pos(iac_pos);
			}

			last_rpm = current_rpm;
			
		} else {  // not in idle-mode

			if (engine.status & _BV(idle)) {
				cli();
				engine.status &= ~_BV(idle);
				sei();

				// this isn't necessary but is nice when logging
				iac.error = 0;
				iac.integral = 0; // this one IS necessary!
				iac.rpmdot = 0;
				iac.pid = 0;
				iac.pid_p = 0;
				iac.pid_i = 0;
				iac.pid_d = 0;
				iac.resync = 10 * config.iac_overclose_interval;
				
				// avoid frying the valve by keeping it powered during non idle
				// lets shut it off, just to be safe.
				if((config.iac_conf & _BV(precise_idle)) && !(config.iac_conf & _BV(stepper_iac))) {
					set_iac_pos(0);
				}
			}
		}
	}
}
#endif



/***************************************************************************/
/*** set iac pos                                                         ***/
/***************************************************************************/
void set_iac_pos(uint8_t pos) {

	if((config.iac_conf & _BV(precise_idle)) && (config.iac_conf & _BV(stepper_iac))) {
		if (!(iac.status & _BV(sync))) {   // don't mess with the sync sequence!

			if (pos < iac.position) {   // close iac
				iac.count = iac.position - pos;
				iac.status &= ~_BV(direction);
				iac.status |= _BV(busy);
			} else {   // open iac
				iac.count = pos - iac.position;
				iac.status |= _BV(direction) | _BV(busy);
			}
		}
	}
	else {
		softpwm_set(&idle_pwm_soft, pos);
		iac.position = pos;
		iac.count = 0;
	}
}



/***************************************************************************/
/*** init iac                                                            ***/
/***************************************************************************/
void init_iac(void) {

// at power on, the iac is fully closed, thereby the 
// position of the iac is known.

#ifdef GENBOARDv3
	// we always enable the chip ?
	cli();
	S259(DHIGH, S259_STEP_EN); 
	sei();
#endif

	if((config.iac_conf & _BV(precise_idle)) && !(config.iac_conf & _BV(stepper_iac))) {
		softpwm_init(&idle_pwm_soft, 0);
		iac.status &= ~(_BV(busy) | _BV(sync));
		iac.position = 0;
		iac.count = 0;
	}
	else {
		iac.status = _BV(busy) | _BV(sync);
		iac.status &= ~_BV(direction);
		iac.position = config.iac_max_steps + IAC_SYNC;
		iac.count = config.iac_max_steps + IAC_SYNC;
	}
}


/***************************************************************************/
/*** move iac                                                            ***/
/*** caller MUST check if stepper_iac is enabled						 ***/
/***************************************************************************/
#define BATT_8V 68

void move_iac(void) {
	static uint8_t phase;

	if (iac.count) {
		// if battery voltage is too low, we don't move the stepper, but just hold 
		// current position to avoid getting out of sync (could be the case 
		// otherwise during cranking)

		if (!(config.iac_conf & _BV(low_power_halt)) || (engine.batt > BATT_8V)) { // battery > 8V?
			uint8_t next_phase;

			iac.count--;
			iac.status &= ~_BV(iac_batt_low);

			if (iac.status & _BV(direction)) {	// open iac
				next_phase = phase + 1; 
				iac.position++; // phase and position does not represent the same number!
			} else {	// close iac
				next_phase = phase - 1; 
				iac.position--;
			}

#ifdef GENBOARDv3
			cli();
			// S259(DHIGH, next_step(next_phase & 0x03));  // activate next wire in sequence
			digitalout((((next_phase << 4) & 0x30) ^ 0x40), DHIGH); // ie. next_phase==3, equivalent mdh70/f0 
			sei_nop_cli();
			// S259(DLOW, next_step(phase & 0x03));  // deactivate previous wire in sequence
			digitalout((((phase << 4) & 0x30) ^ 0x40), DLOW); 
			sei_nop_cli();

			S259(DHIGH, S259_STEP_EN); // this is a bit waste, TODO: find better place to enable the driver
			sei();
#else
			PORTC = next_step(next_phase & 0x03) | _BV(STEP_EN);  // move the iac
#endif

			phase = next_phase;

			if (iac.count == 0) {
				iac.status &= ~(_BV(busy) | _BV(sync));  // when done sync'ing, the position is know
			}
		}/* else {// batt too low, disable driver
		    PORTC &= ~(_BV(STEP_EN) | _BV(STEP_A) | _BV(STEP_B) | _BV(STEP_C) | _BV(STEP_D));
		    iac.status |= _BV(iac_batt_low);
		    }*/
	} else {  // no step requested, disable driver
		iac.status &= ~(_BV(busy) | _BV(sync));
		if (config.iac_conf & _BV(power_off_iac)) {
#ifdef GENBOARDv3
			cli();
			
			digitalout(0x40, DLOW); digitalout(0x50, DLOW); digitalout(0x60, DLOW); digitalout(0x70, DLOW);
			
			S259(DLOW, S259_STEP_EN);
			sei_nop_cli();
			S259(DLOW, S259_STEP_A);
			sei_nop_cli();
			S259(DLOW, S259_STEP_B);
			sei_nop_cli();
			S259(DLOW, S259_STEP_C);
			sei_nop_cli();
			S259(DLOW, S259_STEP_D);
			sei();
#else
			PORTC &= ~(_BV(STEP_EN) | _BV(STEP_A) | _BV(STEP_B) | _BV(STEP_C) | _BV(STEP_D)); 
#endif
		}
	}
}


#ifdef GENBOARDv3
uint8_t PROGMEM STEP_PINS[] = { S259_STEP_A, S259_STEP_B, S259_STEP_C, S259_STEP_D };
#else
uint8_t PROGMEM STEP_PINS[] = { _BV(STEP_A), _BV(STEP_B), _BV(STEP_C), _BV(STEP_D) };
#endif


/***************************************************************************/
/*** next_step                                                           ***/
/***************************************************************************/
// Allow allmost total reconfiguration of the steppers windings connections 
// through PC configuration program. Nice for people too lazy for soldering the
// stepper wires correctly!
uint8_t next_step(uint8_t phase) {
  uint8_t val, idx;

  switch (phase) {
  case 0: idx = config.iac_step_seq & 0x03; break;
  case 1: idx = (config.iac_step_seq >> 2) & 0x03; break;
  case 2: idx = (config.iac_step_seq >> 4) & 0x03; break;
  case 3: idx = (config.iac_step_seq >> 6); break;
  default: idx = 0; break;
  }
  
  val = PRG_RDB(&STEP_PINS[idx]);
  return val;
}

/* ONLY call from userspace!
 * Decide if ON/OFF or PWM strategy is needed. 
 * Call the relevant boolean function depending on strategy. 
 * Than act:
 * check digitalout() for how to configure
 * so DISABLED, p259, stepper or FET can be chosen
 */
void idle_solenoid(void)
{
	uint8_t output;
#ifndef GENBOARDv3
#if ((defined WBO2) || (defined WOT_OUTPUT_TPS_THRESHOLD))
#warning "idle_solenoid not actuated, since WBO2 or WOT_OUTPUT_TPS_THRESHOLD defined"
		return; // on v2 wot output conflicts with fastidle or WBO2
#endif
#endif	// not GENBOARDv3

	if(config.iac_conf & _BV(precise_idle)){
		output = softpwm_act(&idle_pwm_soft);
	} else {
/***************************************************************************/
/*** check fast idle                                                     ***/
/*** Controls a very primitive on/off solenoid to raise idleRPM when cold **/
/***************************************************************************/
	// ths FAST_IDLE_THRES avoids erratic idle when coolant == fastidle
		if (engine.coolant >= config.fastidle) {
			output = 0;
		} else if (engine.coolant < (config.fastidle - FAST_IDLE_THRES)){
			output = DHIGH;		
		} else { // hysteresis window, don't touch the output
			return;
		}
	} // if not precise_idle

	// turns on idle_output if state==DHIGH.
	// turns off idle_output if state==0
	cli();
	digitalout(config.iac_sol_channel, output);
	sei();
} // idle_solenoid()

/*@}*/
