/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen mik at caffrey.dk]        *
***********************************************************
*  multitooth decoding                                    *
\*********************************************************/

#ifndef __MULTITOOTH_C__
#define __MULTITOOTH_C__

#include "global.h"
#include "eventqueue.h"
#include "fuelcontrol.h"
#include "benchmark.h"

#include <inttypes.h>

#define TW_TRIG    0
#define TW_MISSING 1

#define invalidate_rpm_period() engine.rpm_period = 0xFFFFffff
#define isvalid_rpm_period() ((uint16_t)engine.rpm_period) != 0xffff


// true if a > 1.5 * b (NOTE: b2 must be b/2)
#define significantly_bigger(a, b, b2) ((a > b) && ((a - b) > b2))

/*
	Tooth detection.
	Used from trigger interrupt.
  
	Bit definitions of trigger:
    bit 0: trigger
    bit 1: missing tooth
	
	tooth wheel in sync when etime.tooth_cnt < 128
	
	tooth_detection operates on 16 bit period instead of 32 bit
*/
static inline uint8_t tooth_detection(uint16_t period16, uint32_t capturedtime) {
	uint8_t trigger = 0, tooth_cnt, c_trigger_tooth, c_another_trigger_tooth;
	uint16_t etime_atp, etime_atp2;

	tooth_cnt = etime.tooth_cnt;
	
	etime_atp = etime.average_tooth_period;
	etime_atp2 = etime_atp >> 1;	// half
	if (tooth_cnt < 128) {	// wheel_in_sync
		uint8_t ctw = config.tooth_wheel;
		tooth_cnt++;
		if (significantly_bigger(period16, etime_atp, etime_atp2)) { // found the missing tooth
			if (tooth_cnt == ctw) {
				engine.rpm_period = capturedtime - etime.prev_missing_tooth;
				etime.prev_missing_tooth = capturedtime;
				engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
				trigger = _BV(TW_MISSING); // trigger was 0
				tooth_cnt = 0;
#ifdef ALLOW_SLOPPY_TOOTHWHEEL
			// another chance:
			// Marcell's toothwheel sometimes counts 60-3 instead of 60-2
			// this means trigger came later _LAST_TIME_:
			// therefore it _was_ timing 6 degrees later, well..
			// why should we lose the next 2 events ?
			} else if( tooth_cnt == ctw - 1  ){
				engine.rpm_period = capturedtime - etime.prev_missing_tooth;
				etime.prev_missing_tooth = capturedtime;
				engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
				trigger = _BV(TW_MISSING); // trigger was 0
				tooth_cnt = 0;
#ifdef BENCHMARK
				// increase wheel_err so it gets noticed
				benchstats.wheel_err++;
#endif	// BENCHMARK
#endif // ALLOW_SLOPPY_TOOTHWHEEL
			} else {
				// something went wrong, result of the tooth wheel decoding cannot be trusted
				engine.status_trig |= _BV(wheel_error);
				// reenter sync-mode, the only sane thing to do
				tooth_cnt = 255;
				// leave last known good engine.rpm_period as it is
//				engine.rpm_period = 0xFFFF;
#ifdef BENCHMARK
				benchstats.wheel_err++;
#endif	// BENCHMARK
			}
		}
	} else { // NOT wheel_in_sync
#define TOOTH_WHEEL_NOISE 4
#define TOOTH_WHEEL_CNT   4
#define TOOTH_WHEEL_LISTEN (TOOTH_WHEEL_NOISE + TOOTH_WHEEL_CNT)

		engine_is_running(); //turn on fuelpump etc.
// do a rather long sync sequence to avoid getting wild rpm readings during the first rotation

		if (tooth_cnt < (255 - TOOTH_WHEEL_LISTEN)) {
			// initialization done, find the missing tooth
			if (significantly_bigger(period16, etime_atp, etime_atp2)) { // got the tooth
				// uint32_t t32;
				engine.status_trig &= ~_BV(wheel_error);

				// TODO: use timing data from more teeth
				// the >> 1 is because period16 is for 2 teeth at this point
				// would be good for 36-1 only, since 60-2 is 3 teeth !
				// t32 = mult16_8( period16, config.another_trigger_tooth ) >> 1;
				// engine.rpm_period = ((t32 >> 16) ? 65535 : (uint16_t) t32);
				// we just make sure that TW_TRIG not fired in this case:

				invalidate_rpm_period();
	
				etime.prev_missing_tooth = capturedtime;
				trigger |= _BV(TW_MISSING);
				tooth_cnt = 0;
			} 
			// else stay unsynced tooth_cnt kept constant at 254 - TOOTH_WHEEL_LISTEN
			// note that we do update etime.average_tooth_period before return
		} else {  // assume the first pulses are noise
			--tooth_cnt;
			if (tooth_cnt > 255 - TOOTH_WHEEL_NOISE){
				// we don't even avarage at this point.
				// (although a bogus 0xffff * 4usec  would decay fast)
				etime.average_tooth_period = period16;	// doesn't hurt to save it
				etime.tooth_cnt = tooth_cnt;
				return 0;
			} // else note the average_tooth_period averaging
		} 
	}

	// the trigger can be "moved" through config
	// this allows all kind of strange setups, ex. having the missing tooth 180 degrees BTDC
	c_trigger_tooth = config.trigger_tooth;
	c_another_trigger_tooth = config.another_trigger_tooth;
#ifndef TRIGGER_ANOTHER_ANOTHER
#define TRIGGER_ANOTHER_ANOTHER 2
#endif
	if ((tooth_cnt == c_trigger_tooth
		|| tooth_cnt == c_trigger_tooth + c_another_trigger_tooth
		|| tooth_cnt == c_trigger_tooth + c_another_trigger_tooth + c_another_trigger_tooth)
		// you see that this is a hack... it should be a separate
		// downcounter in engine_t that enables several triggers per missing tooth not just 1 or 2
		// or better yet some engphase thingy that supports odd-fire V6
		&& isvalid_rpm_period()
		) {
		trigger |= _BV(TW_TRIG);
		// updating engine.rpm_period and "start a rpm calculation" should come to this point
	}
	// rpmk and ncyl could be used to handle this, but this way we use a more recent rpm_period
	// TODO: more than 4 cyl
	if (c_another_trigger_tooth && 
		(tooth_cnt == c_another_trigger_tooth
		|| tooth_cnt == c_another_trigger_tooth + c_another_trigger_tooth) ) {
		engine.rpm_period = capturedtime - etime.prev_missing_tooth;
		etime.prev_missing_tooth = capturedtime;
		engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
		// trigger = _BV(TW_MISSING); // TODO: do we want this ??
	}

	etime.tooth_cnt = tooth_cnt;
	// first take the half, than add them! Overflow is more concern than 4 usec precision
	etime.average_tooth_period = etime_atp2 + (period16 >> 1);
	return trigger;
}

#endif
