/* Copyright 2005 (C) by Michael Richards. All rights reserved.  */
/* $Id: sipr.c,v 1.6.2.6 2005/05/30 15:35:35 marcell Exp $ */

#include <avr/crc16.h>

#include "ve.h"
#include "inttypes.h"
#include "comm.h"

#include "firmware_revision.h"
#include "sipr.h"

/* global for the current incoming packet */
struct sipr_t sipr_incoming;

/* SIPR 100ms timeout counter for receiving a packet 255 = none incoming */
uint8_t sipr_inc_timeout;
uint8_t sipr_state;

uint8_t sipr_packet_ACK(uint8_t i);
uint8_t sipr_packet_GET_VERSION(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_VE(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_ZERO(uint8_t i);
void sipr_process_packet(void);

/* init function - sets up the timeout and the incoming/outgoing packet data
 */
void sipr_init() {
  /* no packets are incoming */
  sipr_inc_timeout=255;

  /* reset the character count */
  sipr_incoming.idx=0;

  /* reset the sequence number */
  sipr_incoming.seq=0;
}


/* timeout flag. sends a NAK packet */
void sipr_timeout() {
  pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
  sipr_inc_timeout=255;
}


/* process a completed packet from sipr_incoming */
void sipr_process_packet() {
  switch (sipr_incoming.type) {
    case SIPR_PAK_ACK: // got an ACK return an ACK
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq);
      break;
    case SIPR_PAK_GET_VERSION: // return the firmware version
      pushfuncsipr(&sipr_packet_GET_VERSION, 9, sipr_incoming.seq);
      break;
    case SIPR_PAK_SET_BAUD: // return the ack packet and schedule a baudrate change
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq);
      // I'm not sure how I can change the baudrate after this packet has gone out!
      break;
    case SIPR_PAK_GET_TABLESIZE: // return the size of the map tables
      // invalid table ID means we send a zero size
      if (sipr_incoming.data[0]>3) 
        pushfuncsipr(&sipr_packet_GET_TABLESIZE_ZERO, 8, sipr_incoming.seq);
      else {
        if (sipr_incoming.data[0]<=3) pushfuncsipr(&sipr_packet_GET_TABLESIZE_VE, 8, sipr_incoming.seq);
      }
      break;
  }
}

/* processes incoming chars */
void sipr_process_inc(uint8_t c) {
  /* reset the timeout */
  sipr_inc_timeout=2;

  sipr_incoming.idx++;
  switch (sipr_incoming.idx) {
    case 1:
      sipr_incoming.CRC=0;  // don't forget to reset the CRC value because this is a new packet
      sipr_incoming.seq=c;
      break;
    case 2:
      sipr_incoming.type=c;
      break;
    case 3:
      sipr_incoming.size=c;
      break;
    default:
      if (sipr_incoming.idx==sipr_incoming.size+4) {  // CRC byte 1
        sipr_incoming.recCRC=c;
        return;
      }
      if (sipr_incoming.idx==sipr_incoming.size+5) { // CRC byte 2 and check
        sipr_incoming.recCRC |= (c << 8);
        // CRC mismatch send NAK
        if (sipr_incoming.recCRC!=sipr_incoming.CRC) {
          pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
        }
        else { // valid packet
          sipr_process_packet();
        }
        // we are no longer waiting for a packet...
        sipr_inc_timeout=255;  
        return;
      }
      else {
        // accept and store only MAX_INC_SIZE chars
        if (sipr_incoming.idx-4<MAX_INC_SIZE)
          sipr_incoming.data[sipr_incoming.idx-4]=c;
      } 
         
  }  // switch

  /* update the CRC */
  sipr_incoming.CRC=_crc16_update(sipr_incoming.CRC, c);
}


/* NAK packet function */
uint8_t sipr_packet_NAK(uint8_t i) {
  // packet type
  return SIPR_PAK_NAK;
}


/* ACK packet function */
uint8_t sipr_packet_ACK(uint8_t i) {
  // packet type
  return SIPR_PAK_ACK;
}


/* GET_VERSION packet function */
uint8_t sipr_packet_GET_VERSION(uint8_t i) {
  switch (i) {
    case 0: // packet type
      return SIPR_PAK_GET_VERSION;
    case 1:
      return FIRMWARE_VERSION_MAJOR;
    case 2: 
      return FIRMWARE_VERSION_MINOR;
  }
  // case 3
  return FIRMWARE_VERSION_BUILD;
}


/* GET_TABLESIZE packet function for all the VE tables */
uint8_t sipr_packet_GET_TABLESIZE_VE(uint8_t i) {
  switch (i) {
    case 0: // packet type
      return SIPR_PAK_GET_TABLESIZE;
    case 1:
      return VE_SIZE_KPA;
  }       
  // case 2
  return VE_SIZE_RPM;
}


/* GET_TABLESIZE packet function for unknown tables - returns 0's */
uint8_t sipr_packet_GET_TABLESIZE_ZERO(uint8_t i) {
  switch (i) {
    case 0: // packet type
      return SIPR_PAK_GET_TABLESIZE;
  }
  // return 0's for all the sizing data
  return 0;
}
