
/* Copyright 2003
	Authors:
	Marcell Gal cell at x-dsl.hu
	Michael Kristensen mik at caffrey.dk
	Dave Brul
	The WBO2 - different from most other subsystems of the MSAVR project - is
	_not_ GENBOARD PUBLIC LICENSE, but protected by maximum extent 
	of the applicable copyright law.
	No right to distribute, or use in any way unless explicite written permission
	from the authors is granted. 
	It is prohibited to compile the application with WBO2 defined,
	or use part of the copyrighted WBO2 subsystem in any derived work of any nature.
	Just to have it on disk as shipped with the rest of the firmware source is allowed as 
	long as it is not redistributed and not compiled for AVR or any other target system.
	To use it in any way, you must in any case have an explicite written license from the Authors.

	The users who buy WBO2 or GenBoard PCB or kit or assembled circuit 
	from the authors are very likely to be granted the license for using it in the bought
	circuits, but written explicite license is needed.
*/

   /*!
   * \addtogroup g_ego_wbo2 Wide Band O2 Sensor Routines
   */
/*@{*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "wbo2.h"
#include "global.h"
#include "eventqueue.h"
#include "comm.h"

#include "softpwm.c"
#include "adc.h"
#include "pid.h"
#include "helpers.h"
#include "arithm.c"
#include "ego.h"

#include <avr/pgmspace.h>

#ifdef WBO2

struct pid_t heater_pid;
struct pid_t pump_pid;

struct wbo2_t wbo2;
struct wbo2_sampling_t wbo2_sampling;
struct softpwm_t wbo2_heater_pwm;

/* interrupt called functions */
/*
void wbo2_pump_low(void) {
	PUMP_PORT &= ~_BV(PUMP);
}

void wbo2_pump_high(void) {
	PUMP_PORT |= _BV(PUMP);
}
*/

// sets pw in hardware
static inline void pump_actuator(uint8_t pw)
{	
	wbo2.pump_pw = pw;	// save for the curious (logging, LCD)
#ifdef GENBOARDv3
	OCR3A = pw;
#else
	OCR0 = pw;	// set the actuator
#endif
}

static inline void heater_off(void)
{	
	cli();
#ifndef HEATER_INVERTED
	HEATER_PORT &= ~_BV(HEATER);
#else
	HEATER_PORT |= _BV(HEATER);
#endif
	sei();
}

static inline void heater_on(void)
{
	cli();
#ifndef HEATER_INVERTED
	HEATER_PORT |= _BV(HEATER);
#else
	HEATER_PORT &= ~_BV(HEATER);
#endif
	sei();
}

void wbo2_sample_nernst(void) {
	ADMUX = ADMUX_AREF | NERNST; // select nernst channel
	ADCSRA = ( _BV(ADEN) | _BV(ADIF) | _BV(ADIE) | ADCSRA_PRESCALER | ADCSRA_ADFR );
	wbo2_sampling.nernst_sample_cnt = 0;	// start sampling again
// NOTE:
// first interrupt after setting the admux, may contain the result
// of the previous admux value. This is taken care of in the nernst_acdc() processing.

}

// variables for afr filtering
uint8_t afr_filter_idx;
uint16_t afr_history[AFR_FILTER_LENGTH];


static inline void filter_afr(void) {
	uint16_t *p, ca, fa;
	
	if (afr_filter_idx >= AFR_FILTER_LENGTH)
		afr_filter_idx = 0;

	p = &afr_history[afr_filter_idx];
	fa = wbo2.filtered_afr;
	ca = wbo2.current_afr;

	fa -= *p >> 2;
	*p = ca;

	wbo2.filtered_afr = fa + (ca >> 2);
	afr_filter_idx++;
}




// variables for calculating Ri variance
int16_t ri_diff;
int16_t ri_diff_history[WBO2_PUMP_PRESCALE];

static inline uint8_t calc_ri_confidence(void) {
	uint8_t i;
	uint16_t sum, var;

	sum = 0;
	for (i=0; i<WBO2_PUMP_PRESCALE; i++) {
		uint32_t t;

		var = sum;
		t = square_16i(ri_diff_history[i]);
		// why not check t>>16 != 0 instead ? Equivalent with t > 0xffff
		if (t >= 0xffff) {  // overflow will happen, catch it now
			var = 0xffff;
			break;
		}
			
		sum = var + (t & 0xffff);

		if (sum < var) {  // overflow
			var = 0xffff;
			break;
		}
	}

	sum = mult16_8(var, config.wbo2_ri_confidence_scale) >> 16;

	// sum++;   // since FF*FFFF >> 16 = FE, and we would like to
    // be able to have confidence == 0

	if (sum >= 0xfe)
		return 0;
	else
		return 0xfe - sum;
}


// this table is generated by bin/ip2afr.pl
// elements are:  afr*1024
// element distance: 8
// TODO: unfortunately this cannot cover as wide range as the HW supports !!
uint16_t PROGMEM ip_to_afr[] = {
	10407, 10668, 10927, 11188, 11455, 
	11733, 12029, 12352, 12714, 13126, 
	13608, 14177, 14859, 15682, 16681, 
	17897, 19379, 21184, 23381};

// ip_to_afr table is valid in the range dutycycle = [0 ... LAST_ELEMENT]
// for an ideal sensor (eg. no calibration)
#define LAST_ELEMENT 144

// ip_to_afr table has Ip = 0mA at
// Ip = 0mA: duty cycle = 102
#define NEUTRAL_ELEMENT 102

// this table is generated by bin/ip2afr.pl
// elements are:  O2% * 10
// index is Ip=0mA ... Ip=2.62mA
// element distance: 32
uint8_t PROGMEM ip_to_o2[] = { 0, 55, 108, 158, 205, 250, 
// extend a bit so we save a railing:
251,251, 252,252, 253,253, 254,254, 255,255 };

// range might be narrowed in a later revision
static inline uint8_t pump_pw_zero(void)
{
	return config.wbo2_pump_pw_zero;
}

/* 
 Let's do some output conversion for standard logging
 and analog communication (for compatibility)
 Neutral element will map to:
*/
#define WBO2_OUTPUT_STOICH (128)

/* 
 input => output
 32 * 0 => 73.2 (1.43V)
 while stoich 32 * NEUTRAL_ELEMENT
 32 * 102 => 128 (2.5V)
 
 factor: (could be calculated from above)
*/

#define WBO2_OUTPUT_FACTOR (1100)
#define WBO2_OUTPUT_OFFSET (73)
static inline uint8_t Ip_2_wboutput(uint16_t ip)
{
	uint16_t wb16;
	uint8_t wb8;
	
	wb16 = mult16_16(ip, WBO2_OUTPUT_FACTOR)>>16;

	if(wb16 >> 8)
		return 0xff;
		
	wb8 = (uint8_t) wb16;
	
	if(wb8 >= (255-WBO2_OUTPUT_OFFSET))
		return 0xff;
		
	return wb8 + WBO2_OUTPUT_OFFSET;
}

// calculate current afr as a function of pump dutycycle
static inline void calc_afr(void) {
	uint16_t dc_average, k2;
	uint16_t cmd, res;
	uint8_t t8;

	dc_average = wbo2.pump_dc_sum;
//	dc_average = (wbo2.pump_dc_average + wbo2.pump_dc_sum) >> 1;
//	wbo2.pump_dc_average = dc_average;
	// dc_average = WBO2_AFR_SAMPLES * (0..0xff)

    // will this work? What if pump-pid overshoots, then we might be out of the valid range.
    // TODO: calculate variance


/*

The purpose is to calculate res = some_const * dcinput * calmult

calmult multiplier depends on calibration, therefor spans through
(CALIBRATION_BASE .. CALIBRATION_BASE+255) ~ (1 .. 2.54) 
This range seems OK, since

There is the Itotal current through the pump, but 
Bosch uses a theoretical Ip for the AFR table x-axis: 
Ip is a downscale of Itotal, Ip only includes
the measured current flowing through a 61.9 Ohm, 
not including the current through the RCal parallel with 61.9

# in our v3 implementation we use a more precise 
# calibration. The real current generator range span
# 5V / 510 Ohm = 9.8mA over a range of 255 (256?)
# which means 26 command-steps for 1mA 
function out = Realcmd(Itotal)
out=26 * Itotal;
endfunction

# for the ip_to_afr array, the index steps 48 for 1mA change in Ip
function out = ipafr_index(Ip)
out=48 * Ip;
endfunction

function out = RCalmult(Rcalresistor )
temp = 1/61.9 + 1./Rcalresistor;
out = temp .* 61.9;
endfunction

# we want to cover a reasonable part of the 30..300 Ohm LSU4 RCal range:
RCal = [30:20:130,200:100:300]

# [Ip, Ip*ipafr_spacing, 10000, Ip*Calmult(RCal)]

function out = Itotal_Icmd(Ip, RCal)
# res is the output-amplitude, as a result of the calibration
Ip2afr_res = ipafr_index(Ip)
Itotal= Ip .* RCalmult(RCal);
# from Itotal we calculate the OCRx command amplitude,
# the input of the calibration:
dc = Itotal .* Realcmd(1);
out = [RCal; Itotal ; dc ];
endfunction

Ip = 0.32673
Itotal_Icmd(Ip, RCal)

Ip = 0.8
Itotal_Icmd(Ip, RCal)

Ip = 1.0
Itotal_Icmd(Ip, RCal)

# from the octave result, you can see that 
# if the 26 and 48 constants above are OK,
# we want the (1 .. 2.54) range as *(0.6027 .. 1.5304)

---------------------------
The sugar in the calc is caused by the fact that
dcinput is biased at pump_pw_zero() (call p) 
while output is biased at NEUTRAL_ELEMENT (call 102 :-)

Octave snippet doing the same calibration function:

dutycycle = 0:1:255;
pump_pw_zero = 107;
calibration = 150;

# This is the real thing:
function y = f(c, d, p)
CALIBRATION_BASE=166;
WBO2_MAGIC_MULT=1904;
OUTPUT_BIAS_SCALEDUP=102*32/WBO2_MAGIC_MULT*256
k2 = c + CALIBRATION_BASE;
y = (k2 * (4 * d + 4*(256 - p))/256 + OUTPUT_BIAS_SCALEDUP - 4*k2) * WBO2_MAGIC_MULT/256;
endfunction

a=[dutycycle; f(calibration, dutycycle, pump_pw_zero)/32]

grid "on"
plot(dutycycle, f(calibration, dutycycle, pump_pw_zero))

example of ideal reference curve vs. example calibration:
plot(dutycycle, f(128, dutycycle, 102), dutycycle, f(171, dutycycle, 107))
note that f(128, dc in free air, 102) == f(cal, dc in free air, offset)

If you understand all the above, you see where these beauties come from:
*/
#define CALIBRATION_BASE (166)
#define WBO2_MAGIC_MULT (1904)
#define OUTPUT_BIAS_SCALEDUP (439)

	k2 = ((uint16_t)config.wbo2_calibration) + CALIBRATION_BASE;

	// let's move the bias (Ip=0mA) to 256 (before *= 4)
	t8 = 0 - pump_pw_zero(); // 256 - ... as you see
	// <<2 because of WBO2_AFR_SAMPLES
	cmd = ((uint16_t)(t8)) << 2 ;
	cmd += dc_average;	// already 4 values summed
	// == [0 : 408: 1428 ]
	res = mult16_16(cmd, k2) >>8;

	// bias so the output will be biased just as we like:
	res += OUTPUT_BIAS_SCALEDUP;
	
	k2 <<= 2;
	// capture negative Ip overflow
	if (k2 > res) {
		//res = 0;   // a slightly wrong calculation is better than a disaster!
		// TODO: flag that the wbo2 is railed at a maximum rich condition
		wbo2.current_afr = 0xffff;
		return;
	}
	res -= k2;
	
 	res = mult16_16(res, WBO2_MAGIC_MULT)>>8;

	// we could maybe assign OC3C to have a HW output?
	wbo2.Ip_output = Ip_2_wboutput(res);
	
	// please review the amplitude of res !!!
	// should be 32*(102+signed_Ip) == 32 * tbl-index

	if (res <= (LAST_ELEMENT << 5) ) { // check if idx is in range for calculation AFR
		// interpolate in ip_to_afr table

		uint16_t a, b;
		uint8_t idx, offset;

		idx = (uint16_t)res >> (5+3);  // elements in the table has distance 8 (3 bits)
		offset = (uint16_t)res & 0xff;

		a = pgm_read_word(&ip_to_afr[idx]);
		b = pgm_read_word(&ip_to_afr[idx+1]);

		// table ip_to_afr[]is increasing
		wbo2.current_afr = a + (mult16_8(b - a, offset) >> 8);
		wbo2.current_o2 = 0;

		filter_afr();

		// TODO: the confidence should also depend on the pump dutycycle (use filtered_afr).
		wbo2.afr_confidence = calc_ri_confidence();

	} else {
		// not in the lambda 0.6 ... 1.5 range
        // calculate O2% instead
		
		uint16_t t;
		uint8_t idx, offset, a, b;

        // table is only valid for Ip >= 0mA
		// since the above if-statement was false, res > LAST_ELEMENT<<5
		// and LAST_ELEMENT > NEUTRAL_ELEMENT ==> Ip >= 0mA

		// the largest possible value of res = 2*32D9 (15 bits)
		// the upper 3 bits used for idx (only idx 0..5 actually used)
		// the next 8 bits for offset, rest are thrown away
		t = res - (NEUTRAL_ELEMENT << 5);
		t >>= 2;
		idx = t >> 8;  // elements in the table has distance 32
		offset = (uint8_t)t; // lower byte of t

		if (idx <= 4) {

			a = pgm_read_byte(&ip_to_o2[idx]);
			b = pgm_read_byte(&ip_to_o2[idx+1]);

			// table ip_to_o2[]is increasing
//			wbo2.current_o2 = a + (((b - a) * offset) >> 8);
			wbo2.current_o2 = ((uint16_t)a<<8) + (b - a) * offset;
		} else {
			// something went wrong, maybe pump_pw_zero is set way low?
			wbo2.current_o2 = 0;
		}

		// set a status flag somewhere to indicate that the wbo2.current_afr
		// cant be used for ego control.
		wbo2.current_afr = 0xffff;
	}
}


/* 
Unfortunately we must take into consideration
when the nernstpulse edge happens:
edgetime[n]=the read time (TCNT2) of the edge before 
the sample[n+1], 0..255 [0.5usec].
between the 2 samplings, because of the discharge of C12:
(linear approximation of the exponential decay)
upto 104usec * 2* 250uA / 220 nF = 240mV 
-> this is small /~5%/ compared to 5V,
but not insignificant. A simple correction with the
edgetime will be perfect:
if edgetime is bigger, the _seen_ ADC jump absolute value will
be smaller, so positive correction is needed.
Returned in ADC LSB units (normally very small, appr. 0..40)
*/
static inline uint8_t edgetime_corr(uint8_t et)
{
	uint8_t r, corr;
/*
Uc = (-2 * edgetime * 250uA / 220nF) * Ri/(Ri+10k)
(let the edgetime multiplier be config.edgetime_corr,
which includes approximate 1/(Ri + 10k).
Here unfortunately Ri is not yet known exactly, but we can
use an estimated value (calculated without the Uc correction, thus
including some error) or rather previous Ri is fine and simple:
we'll use that method.

amplitudes: 220nF and 10k assumed, Ri=80 (and Ri2=100 in parenthesis) Ohm, 
note: these could be octave or matlab formulas..
pulsing = 5V
pulsing on the Ri: 5V * Ri/(10k+Ri) = 39.68 (49.5) mV
after the nernst OPA amplification: 6x and ADC 5000mV/1024
 49 (61) ADC-counts
 wbo2.Ri (sum of 8 of the above:) 390 (487)
 tau = RC = 220nF * 10080 Ohm = 2.218 msec, t=104 usec
 exp(-t/tau) = exp(-0.049) = 1-0.049 => result should be about 2*0.049 of the ADC-counts: 4.77 (5.94)
 when et == 208 (104 usec in 1/2 usec resolution)
 
 TODO: recalculate this.
 suggested config.edgetime_corr:0xBA
 see
 http://www.squirrelpf.com/msavr/index.php?page=WideBandO
 Note, that OPA amplification should not change config.edgetime_corr,
 since it changes the jumps and the necessary additive 
 correction with the same factor it changes the wbo2.Ri value.
 Because of the offset-cheat, maybe r = (wbo2.Ri>>4) - config.whatever; 
 would make a tiny difference for the better.

*/
	r = wbo2.Ri >> 4;
	corr = (r * config.wbo2_edgetime_corr) >> 8;
	return (et * corr)>>8;
}

// how many of the biggest jumps should we care?
#define TOP_N 3
void find_max(uint16_t d, uint16_t *diffs, uint8_t *index, uint8_t s_idx, uint8_t edgetime)
{
	uint8_t i, b;
	/* if bigger than diffs[0] we put it into the array, moving the elements down one index */
	if(d > diffs[0]){	// big enough, insert into the arrays
	
	// cheat1: we only want to run correction for the highest N jumps, not all
		
	// Note1: overflow of the TCNT2 will not cause problem, because when it happens
	// there could be no jumps in the samples, so discarded anyway
	
	// Note2: the sampling time was stored earlier than the sample itself (therefore the -1)
	
	// cheat2: instead of -= (subtracting), and calculating with (104usec-edgetime) we += (add)
	// this will cause an offset, thus Ri=0 Ohm will not be zero, but offsetted. No problem.

		d += edgetime_corr(edgetime); 
		b = 1;
		// find the index after insertion
		while((d > diffs[b]) && (b < TOP_N)){	
			b++;
		}

		b--;	// go back a little

		for(i=0; i < b; i++){	// shift the smaller ones down
			index[i] = index[i+1];
			diffs[i] = diffs[i+1];
		}

		// yeah, insert it
		index[b] = s_idx;
		diffs[b] = d;
	} // if 
}

// filter (sum), check for overflow
void nernst_filt(uint16_t Ri, uint16_t dc)
{
	uint16_t t;
#ifdef BENCHMARK
	uint16_t b;
#endif
	t = wbo2.Ri_sum;
#ifdef BENCHMARK
	b = t;	// backup
#endif
	t += Ri;
#ifdef BENCHMARK
	if(t < b){	// oops, overflow!
		benchstats.wbo2_overflow_Ri++;
		t = 0xffff;
	}
#endif
	wbo2.Ri_sum=t;

	t = wbo2.nernst_dc_sum;
#ifdef BENCHMARK
	b = t;	// backup
#endif
	t += dc;
#ifdef BENCHMARK
	if(t < b){	// oops, overflow!
		benchstats.wbo2_overflow_nernst++;
		t = 0xffff;
	}
#endif
	wbo2.nernst_dc_sum = t;;
}

/* ------------------ */
static inline void check_sampleidx_range(uint8_t i)
{
	return; // I dont care, my nerves can't handle seing wheel_err
		// changing. Find another variable to bother.

#ifdef BENCHMARK
	// i-1 must be in range, but actually, i==1 is suspicious too,
	//	because 1st sample is not from nernst (another ADC channel)
	if(i <= 1 || i >= WBO2_SAMPLE_COUNT){	// oops, BAD
		benchstats.wheel_err++;	// argh, benchstats is full!
	}
#endif
}
/* -------------------- */
/* goes through the samples and finds big jumps (samplepairs).
Applies edgetime correction, outlier elimination and calculates
the AC (Ri) and DC components from the samplepairs.
*/
void nernst_acdc(void)
{
	uint8_t i, edgetime;
	uint16_t sum, dsum;
	// biggest up and downjumps are sorted to separate arrays
	uint16_t udiffs[TOP_N], ddiffs[TOP_N];	// 0 is smallest, last is biggest jump
	uint8_t uindex[TOP_N], dindex[TOP_N];	// index to the original array: nernst_sample[WBO2_SAMPLE_COUNT]

	uint8_t *comm_TX_plainbytesP;
	uint16_t t16;
	
	// initialize the array
	i = TOP_N;
	do {
		i--;
		udiffs[i] = ddiffs[i] = 0;
		// indexes need to be initialized, cause otherwise all 0 samples could screw us
		uindex[i] = dindex[i] = 1;	// 1 is the smallest valid nernst sample index
	} while(i);
/* do we want quicksort or heapsort? 
* Now we go through the samples and keep the TOP_N biggest upjumps and TOP_N biggest downjumps
*/
	for(i=2; i < WBO2_SAMPLE_COUNT; i++){
		uint16_t d1, d2;
	
		d2 = wbo2_sampling.nernst_sample[i];
		d1 = wbo2_sampling.nernst_sample[i-1];
		edgetime = wbo2_sampling.nernst_edgetime[i - 1];
		
		// because of the inductivity of the stupid sensor wires, we MUST
		// neglect the samples taken in the 0..25usec window after pulsing
		if(edgetime > config.wbo2_edgetime_min){ // TODO: do we always have enough samples?
		
		if(d2 > d1) // upjump
			find_max(d2 - d1, udiffs, uindex, i, edgetime);
		else	// downjump
			find_max(d1 - d2, ddiffs, dindex, i, edgetime);
			
		}
		/* calculate the jump */	

	} // for
	
	/* only use [0] and [1] for ACDC calc, eliminate the highest ones ([2] and [3]) */

	/* the selected upjumps */
	i = uindex[0]; check_sampleidx_range(i);
	sum = wbo2_sampling.nernst_sample[i] + wbo2_sampling.nernst_sample[i - 1 ];
	i = uindex[1]; check_sampleidx_range(i);
	sum += wbo2_sampling.nernst_sample[i] + wbo2_sampling.nernst_sample[i - 1 ];
	dsum = udiffs[0] + udiffs[1];
	
	/* the selected downjumps */
	i = dindex[0]; check_sampleidx_range(i);
	sum += wbo2_sampling.nernst_sample[i] + wbo2_sampling.nernst_sample[i - 1 ];	
	i = dindex[1]; check_sampleidx_range(i);
	sum += wbo2_sampling.nernst_sample[i] + wbo2_sampling.nernst_sample[i - 1 ];
	dsum += ddiffs[0] + ddiffs[1];
	// update wbo2...
	wbo2.Ri = dsum;		// scale: sum of 4 jumps
	wbo2.nernst_dc = sum;	// scale: sum of 8 ADC samples

	nernst_filt(dsum, sum);

	// let's log the wbo2 samples and results
	if((debug.enable_flags & _BV(0))	// wbo2 sample-logging enabled
	&& (*(comm_TX_plainbytesP = comm_TX_plainbytes) == 0)	// the signalling byte is zeroed by comm.c (gcc-at-powerup), great
	&& (cmd_queue_head != ((cmd_queue_tail + 1) % CMD_QUEUE_SIZE))) {	// got some place in cmd buffer
	
		*comm_TX_plainbytesP++='\r';	// the same char as checked in above condition
		*comm_TX_plainbytesP++='\n';
		*comm_TX_plainbytesP++='w';
		*comm_TX_plainbytesP++='=';
	
		for(i=0; i < WBO2_SAMPLE_COUNT; i++) {
		/* somewhere, maybe comm.h / comm.c ?
		
		#define comm_TX_plain_SIZE (64)
		and declare
		uint8_t comm_TX_plainbytes[comm_TX_plain_SIZE]
	
		log (maybe using ringbuf, but you can just do it from here, as you like)
		the wbo2_sampling.nernst_sample[i-1]; and  wbo2_sampling.nernst_edgetime[i-1] 
		as binary values (2+1 bytes) into that buffer.
		Put some marker to the beginning like "\nw="
		At the end, call 
		pushfuncpar2(srl_generic, number_of_bytes_to_send, pointer_to_byte0 )
		*/
			t16 = wbo2_sampling.nernst_sample[i];
			*comm_TX_plainbytesP++ = t16 >> 8;
			*comm_TX_plainbytesP++ = (uint8_t)(t16);
			*comm_TX_plainbytesP++ = wbo2_sampling.nernst_edgetime[i];
		}
		t16 = wbo2.Ri;
		*comm_TX_plainbytesP++ = t16 >> 8;
		*comm_TX_plainbytesP++ = (uint8_t)(t16);
		t16 = wbo2.nernst_dc;
		*comm_TX_plainbytesP++ = t16 >> 8;
		*comm_TX_plainbytesP++ = (uint8_t)(t16);
		t16 = heater_pid.integral;
		*comm_TX_plainbytesP++ = t16 >> 8;
		*comm_TX_plainbytesP++ = (uint8_t)(t16);
		t16 = pump_pid.integral;
		*comm_TX_plainbytesP++ = t16 >> 8;
		*comm_TX_plainbytesP++ = (uint8_t)(t16);
		*comm_TX_plainbytesP++ = wbo2_heater_pwm.on_ratio;

		pushfuncpar(NULL, WBO2_SAMPLE_COUNT * 3 + 4 + 9, &comm_TX_plainbytes[0]);
		// pushfuncpar(NULL, 26, "abcdefghijklmnopqrstuvwxyz");
		/* do the ringbuf_push_n after this */
	}
	
}


//#ifndef PUMP_PW_ZERO
// initial pw: lo=-2mA hi=+3mA, we want 0mA which is appr. 255*2/5 = 102
//#define PUMP_PW_ZERO (104)
//#endif
/*
  ocr2 period = 512 usec
  time to do one adc conversion = 16e6 / 128 * 13 + interrupt overhead = 104 usec + int. overhead
*/


void init_wbo2(void) {
	wbo2.status = 0;
	heater_off();
	etime.wbo2_sampling = WBO2_SAMPLING+1;

	// otherwise ADIE should be enabled 
	// Note: without ADIE with nernst_sample_cnt=0, sampling would be disabled!
	wbo2_sampling.nernst_sample_cnt = 0xff;
	
#ifdef GENBOARDv3
	// setup timer3 16uS period and turn on pwm mode
	TCCR3A =  _BV(COM3A1) | _BV(WGM30); 
	TCCR3B =  _BV(WGM32) | _BV(CS30);
#else
	// setup timer0 16uS period and turn on pwm mode
	TCCR0 =  _BV(WGM00) | _BV(COM01) | _BV(WGM01) | _BV(CS00); 
#endif

	pump_actuator(config.wbo2_pump_pw_zero);
	
	// setup timer (OC2)
	// OCR2 50% square, 1kHz...4kHz
//	TCCR2 = _BV(OCIE2) | _BV(WGM21) | _BV(CS21) | _BV(CS20);
	TCCR2 = _BV(WGM21) | _BV(COM20) | _BV(CS21);

	// ocr2_freq = 16e6 / (2*64*(1+ocr2))
	//	OCR2 = 63; // 1953.125 Hz (512 usec period)
#ifdef HACK
	OCR2 = config.ego_target;	// serious hack
#else
	OCR2 = 0xff; // 3906.25 Hz (256 usec period)
#endif

	pid_init(&heater_pid, (struct pid_parameters_t*) &config.wbo2_heater_pid_kp);
	pid_init(&pump_pid, (struct pid_parameters_t*) &config.wbo2_pump_pid_kp);
	
	// should we call these every second? Reboot after changing them
	// OPA amplification, g = 270/75 + 1 = 4.6

	// 80 Ohm / 10k is 1/125 divider, which is g*125 after OPA amplification
	// because of the number of samples, ADC precision and shifts,
	// 4096*g/125 = 151 = 0x97 is a good config.wbo2_ri_target

#define RI_INTERNAL_TARGET (((uint16_t)config.wbo2_ri_target)<<3)
	pid_set_target(&heater_pid, RI_INTERNAL_TARGET);// 2(4 samples) +2 bit ADC, -4: see scale in config_t!

	// Vadc = g*(5-Vnernst)
	// Vnernst = 4.45V ==> Vadc = 2.53V
	// nernstdc_target = Vadc/g * 256 = 141 = 0x8D

	wbo2_pump_pid_set_target(&pump_pid, config.wbo2_nernstdc_target);

	// init these, else the display will show funny things
	wbo2.current_afr = 0xffff;
	wbo2.current_o2 = 0;
}

/*
temperature, starting voltage (from datasheet "standard")
-40C, 7.4V
-10C, 7.8V
 20C, 8.2V
 50C, 8.6V
dC=30, dV=0.4 ... great!

dutycycle at -40 C == -40 F (engine.coolant==0)   : 7.4/(13/256) = 146
dutycycle at  50 C == 122 F (engine.coolant==161) : 8.6/(13/256) = 169
*/

#ifndef HEATER_START_COLD
// conservative low-power starting: 6..7 V
#define HEATER_START_COLD ((uint8_t)(6*6*256/30/30))
#define HEATER_START_WARM ((uint8_t)(7*7*256/30/30))
#endif

#ifndef WARM_TEMP
#define WARM_TEMP (161)
#endif
// interpolate the starting voltage
static inline uint8_t heater_initial_dc(void) {
	if (engine.coolant >= WARM_TEMP) return HEATER_START_WARM;
#define FACTOR ((uint8_t)((HEATER_START_WARM - HEATER_START_COLD)*256/WARM_TEMP))
	return ((FACTOR * engine.coolant) >> 8) + HEATER_START_COLD;
}

// we don't return ws new value just save it globally
static inline void leave_warmup(uint16_t now_upper, uint8_t ws)
{
	ws &= ~_BV(IN_WARMUP);
	ws |= _BV(PUMP_ENABLE);  // make sure that the pump is enabled
	wbo2.status = ws;
	// the Ri-offtarget limit should not interfere with WARMUP, we
	// start fresh after leaving WARMUP
	wbo2.heater_railed_since = now_upper;

	// TODO: We can enable O2 earlier than this
	cli();
	engine.status_ext &= ~_BV(o2_not_ready);  // enable the O2
	sei();

	// do we want to reset wbo2 heater integral ?
}

// used to limit heating at rampup and watch for railing
// Ip controller jumped in here to reuse WARMUP branch
static inline uint8_t heater_limit(uint8_t heat)
{
	uint16_t dc, now_upper, tdiff, maxtime, whrs;
	uint8_t limit, ws, wsf;
	
	now_upper = etime.now_upper;
	ws = wbo2.status;		// early writeback (rarely changed)
	if (ws & _BV(IN_WARMUP)) { // ramp heater voltage

		// start pump pid controlling earlier if Ri gets on target
		if (wbo2.Ri_sum <= RI_INTERNAL_TARGET) {
			leave_warmup(now_upper, ws);
			// no return ! limiting "heat" is still applied below
		}

		// note, if wbo2_warmup_ramp == 0, warmup mode will never be left
		dc = mult16_8(now_upper - wbo2.started, config.wbo2_warmup_ramp) >> 8;
		dc += (uint16_t)wbo2.heater_start_dc;

		// think of the amplitude of 'limit': with ADC0 railing at 30V and
		// softpwm_act_var(&wbo2_heater_pwm, bat*bat)
		// 12V is only 12*12 / 30/30 * 256  => limit==41

		if (dc >> 8){	// leave WARMUP
			leave_warmup(now_upper, ws);
			return heat;
		} else {
			limit = (uint8_t)dc;
			// wbo2_warmup_target was used to terminate the rampup before wbo2_abs_limit (bad idea)
			// wbo2_warmup_target is DROPPED
			if ( limit >= config.wbo2_abs_limit ) {
				// leave WARMUP
				// otherwise we would go above config.wbo2_abs_limit
				leave_warmup(now_upper, ws);
				limit = config.wbo2_abs_limit;
			}

			if(heat > limit) heat = limit;

			// we return. Note that wbo2_abs_limit was also applied
			return heat;
		}
		
		// no need to repeat: above both branches return heat;
	} 
	
	// not in WARMUP

	wsf = ws & _BV(FALLBACK_ACTIVE);
	if(wsf){	// FALLBACK_ACTIVE
		limit = config.wbo2_fallback;
		maxtime = config.wbo2_retry_t << 4;	// appr. 4 sec, not 262msec
	} else {	// no FALLBACK, using wbo2_abs_limit
		limit = config.wbo2_abs_limit;
		maxtime = config.wbo2_limit_maxt;	// 262 msec
	}

	whrs = wbo2.heater_railed_since;	// late writeback, since written back in most cases

	if(heat > limit){	// we are railing...
		heat = limit;	// sorry, you're not allowed to be this high!
		tdiff = now_upper - whrs;	// railing for how long time now?
		if (tdiff > maxtime) { // we've spent too much time at railing
			// if we were in fallback, try control again. If we were at abs_limit, go to fallback
			ws ^= _BV(FALLBACK_ACTIVE); // never conflicts with status change of leave_warmup()
			wbo2.status = ws;
			whrs = now_upper;
		}
	} else {	// not railing, give it more time in this state
		// we don't set wbo2.heater_railed_since to current time if FALLBACK_ACTIVE :
		// even if controller succeeds for moments to control without railing:
		// we'll always try again after config.wbo2_retry_t elapsed
		if(wsf == 0) whrs = now_upper;
		// this small extra condition evaluation could be eliminated by duplicating 
		// the above 10 lines for the previous FALLBACK_ACTIVE branches...
	}

	wbo2.heater_railed_since = whrs;
	return heat;
}

static inline void set_pump(void) {
	uint8_t t, ws;

	ws = wbo2.status;

	if (ws & _BV(PUMP_ENABLE)) {

		if (debug.enable_flags & _BV(4)) {
			// only for testing !
			t = sensors8(MAP);
			pump_actuator(t);
		} else {
#ifdef BENCHMARK
		// peak in the pid parameters and detect (+log) overflow
		// this test will go away when it is verified that the calculation doesn't overflow
		uint8_t a, b;
		int16_t res;
		a = pump_pid.target >> 8;
		b = wbo2.nernst_dc_sum >> 8;
		res = (uint16_t)a - (uint16_t)b;
		if ((res > 127) || (res < -128))
			benchstats.ocr0cnt++;
#endif
		t = add8_16idiv16(config.wbo2_pump_pw_zero, pid15(&pump_pid, wbo2.nernst_dc_sum>>1));	

// These values serves to prevent a _very_ oscillating pump behavior during sensor warmup.
// When the sensor gets on target first time, the pump is enabled. This somehow makes the 
// sensor Ri measurement go skyhigh and requires even more power to the heater. While heater
// is working getting the sensor on target (again), the max pump current is limited.
#define SENSOR_COLD_MAX_DPUMP 64
#define SENSOR_COLD_MIN_DPUMP 32

		if (wbo2.Ri > (config.wbo2_ri_target<<2)) {	// Ri too high, let's try not to kill the sensor
			// TODO: if pump_pw_zero is set to a very low value, an overflow occurs
			// BUT: pump_pw_zero is scheduled to get redefined with a more narrow operating range
			t = limit_u8(t, config.wbo2_pump_pw_zero+SENSOR_COLD_MAX_DPUMP,
				     config.wbo2_pump_pw_zero-SENSOR_COLD_MIN_DPUMP);
		}

		pump_actuator(t);

		}

		wbo2.pump_dc_sum += t;  // used for the afr calculation
		// should each of the t's be stored in an array instead for further calculations?

		wbo2.afr_samples--;
		// store the Ri difference
		ri_diff_history[wbo2.afr_samples] = ri_diff;

		if (wbo2.afr_samples == 0) {
			wbo2.afr_samples = WBO2_AFR_SAMPLES;
			calc_afr();
			// note that this is a bit rare, but why would we call it more
			// frequent than calc_afr ?
			if(WBO2_SENSOR_CONFIGURED)
				o2();

			wbo2.pump_dc_sum = 0;
		}

	}
}

// this function takes care of warmup and the low level heater control
void wbo2_heater(void) {
	uint8_t ws;
	ws = wbo2.status;	// cached: written back after each change
	if ((engine.status & _BV(running)) || (debug.enable_flags & _BV(1))) {  // engine is running,

		if ((ws & _BV(ACTIVE))==0) {
			// first run, initialize variables
			wbo2.started = etime.now_upper;
			wbo2.heater_start_dc = heater_initial_dc();
			softpwm_init(&wbo2_heater_pwm, 0);
			ws |= _BV(ACTIVE) | _BV(IN_WARMUP);
			wbo2.status = ws;
			pump_actuator(config.wbo2_pump_pw_zero);

			wbo2.heater_prescale = WBO2_HEATER_PRESCALE;
			wbo2.afr_samples = WBO2_AFR_SAMPLES;
			wbo2.Ri_sum = 0;
			wbo2.nernst_dc_sum = 0;
			wbo2.pump_dc_sum = 0;

			wbo2.current_afr = 0xffff;
			wbo2.current_o2 = 0;
			ri_diff = 0;
		}
		
		// if ==ff, we already evaluated (don't evaluate again...)
		if(wbo2_sampling.nernst_sample_cnt == WBO2_SAMPLE_COUNT){	// evaluate nernst samples and run pid
			uint8_t heat;

			nernst_acdc();
			// no need to cli(), since adc irq will not change it in this state
			wbo2_sampling.nernst_sample_cnt = 0xff;	// flag that evaluation is done
			// TODO: double check handler2k and confirm that the above will not 
			// corrupt memory in adc interrupt
		
			if (--wbo2.heater_prescale == 0) {
				uint8_t a, b;
				int16_t res;

				wbo2.heater_prescale = WBO2_HEATER_PRESCALE;

				// this branch is run every 8 (or so) msec, so good for calling pid	
				// too high Ri wants more heat! (higher T => lower Ri)

				// in warmup we don't want the pid to interfere
				if (ws & _BV(IN_WARMUP))
					heat = config.wbo2_warmup_target;
				else {
#ifdef BENCHMARK
					// peak in the pid parameters and detect (+log) overflow
					// this test will go away when it is verified that the calculation doesn't overflow
					uint8_t a, b;
					int16_t res;
					a = heater_pid.target >> 8;
					b = wbo2.Ri_sum >> 8;
					res = (uint16_t)a - (uint16_t)b;
					if ((res > 127) || (res < -128))
						benchstats.ocr0cnt += 0x0100;

#endif
					heat = sub8_16idiv4(config.wbo2_fallback, pid15(&heater_pid, wbo2.Ri_sum));// higher byte of u16 loaded first (finally the lower)
				}

				heat = heater_limit(heat);	// changes wbo2.status !
				// ws = wbo2.status;	// not needed, since the only bit that we care _BV(ACTIVE) never changed in heater_limit()
				softpwm_set(&wbo2_heater_pwm, heat);
				// softpwm_set(&wbo2_heater_pwm, 0x1c);
				wbo2.heat = heat;

				// calculate Ri difference (for calculating afr confidence)
				a = wbo2.Ri_sum >> 8;
				b = RI_INTERNAL_TARGET >> 8;
				res = (uint16_t)a - (uint16_t)b;
				if (res > 127) 
					ri_diff = 32767;
				else if (res < -128) 
					ri_diff = -32768;
				else
					ri_diff = wbo2.Ri_sum - RI_INTERNAL_TARGET;


				wbo2.Ri_sum = 0;
			}

			if (--wbo2.pump_prescale == 0) {
				wbo2.pump_prescale = WBO2_PUMP_PRESCALE;

				set_pump();
				wbo2.nernst_dc_sum = 0;
			}
		}
		
		// looks like this will always be true, because we set the flag above: (?)
		if (ws & _BV(ACTIVE)) {
			uint8_t bat;
			bat = engine.batt;
			// TODO: check if set in calc_parameters and do scaling, and softpwm_act_var() is...
			// (bat*bat - config...) would be superfluous: the heater FET is ohmic, not voltage drop
			// the heater is pwm driven
			if(softpwm_act(&wbo2_heater_pwm)){
				heater_on();
			} else {
				heater_off();
			}
		}

	} else {	// engine not running
		// note that the NERNST sampling is kept disabled! should not 
		// seriously affect the other sensors!!!
		// wbo2.status &= ~_BV(NERNST_READY);
		if (ws & _BV(ACTIVE)) {
			// the engine just stopped
			heater_off();
			pump_actuator(config.wbo2_pump_pw_zero);			
			wbo2.status = 0;

			// used by lcd display:
			wbo2.heat = 0;
			wbo2.Ri = 0;
			wbo2.nernst_dc = 0;

			wbo2.current_afr = 0xffff;
			wbo2.current_o2 = 0;

			cli();
			engine.status_ext |= _BV(o2_not_ready);  // disable the O2
			sei();
	
		}
	}
}	// wbo2_heater()

#endif

/*@}*/

